# -*- coding: utf-8 -*-

"""
This module contains utilities for support Zipkin distributed tracing system:

http://twitter.github.io/zipkin/
"""
from __future__ import absolute_import

from pyramid_zipkin.request_helper import create_zipkin_attr
from pyramid_zipkin.thread_local import get_zipkin_attrs
from pyramid_zipkin.thrift_helper import create_endpoint
from pyramid_zipkin.thrift_helper import generate_span_id
from pyramid_zipkin.logging_helper import ZipkinLoggerHandler
from pyramid_zipkin.logging_helper import ZipkinLoggingContext


def zipkin_tween(handler, registry):
    """
    Factory for pyramid tween to handle zipkin server logging.

    :param handler: pyramid request handler
    :param registry: pyramid app registry

    :returns: pyramid tween
    """
    def tween(request):
        zipkin_attrs = create_zipkin_attr(request)
        endpoint_attrs = create_endpoint(request)
        log_handler = ZipkinLoggerHandler(zipkin_attrs)
        with ZipkinLoggingContext(zipkin_attrs, endpoint_attrs, log_handler,
                                  request) as context:
            response = handler(request)
            context.response_status_code = response.status_code
            return response

    return tween


def create_headers_for_new_span():
    """
    Generate the headers for a new zipkin span.

    .. note::

        If the method is not called from within a pyramid service call OR
        pyramid_zipkin is not included as a pyramid tween, empty dict will be
        returned back.

    :returns: dict containing (X-B3-TraceId, X-B3-SpanId, X-B3-ParentSpanId,
                X-B3-Flags and X-B3-Sampled) keys OR an empty dict.
    """
    zipkin_attrs = get_zipkin_attrs()

    if not zipkin_attrs:
        return {}

    return {
        'X-B3-TraceId': zipkin_attrs.trace_id,
        'X-B3-SpanId': generate_span_id(),
        'X-B3-ParentSpanId': zipkin_attrs.span_id,
        'X-B3-Flags': '0',
        'X-B3-Sampled': '1' if zipkin_attrs.is_sampled else '0',
    }



# -*- coding: utf-8 -*-
import os
import socket
import struct

import thriftpy
from thriftpy.transport import TMemoryBuffer
from thriftpy.protocol.binary import TBinaryProtocol

from pyramid_zipkin.request_helper import generate_span_id


thrift_filepath = os.path.join(os.path.dirname(__file__),
                               'thrift/zipkinCore.thrift')
zipkin_core = thriftpy.load(thrift_filepath, module_name="zipkinCore_thrift")


dummy_endpoint = zipkin_core.Endpoint()


def create_annotation(timestamp, value, host):
    """
    Create a zipkin annotation object

    :param timestamp: timestamp of when the annotation occured in microseconds
    :param value: name of the annotation, such as 'sr'
    :param host: zipkin endpoint object

    :returns: zipkin annotation object
    """
    return zipkin_core.Annotation(timestamp=timestamp, value=value, host=host)


def create_binary_annotation(key, value, annotation_type, host):
    """
    Create a zipkin binary annotation object

    :param key: name of the annotation, such as 'http.uri'
    :param value: value of the annotation, such as a URI
    :param annotation_type: type of annotation, such as AnnotationType.I32
    :param host: zipkin endpoint object

    :returns: zipkin binary annotation object
    """
    return zipkin_core.BinaryAnnotation(
        key=key, value=value, annotation_type=annotation_type, host=host)


def create_endpoint(request):
    """Create a zipkin endpoint object based on a request

    :param request: pyramid request object
    :returns: zipkin endpoint object
    """
    service_name = request.registry.settings['service_name']
    host = socket.gethostbyname(socket.gethostname())
    port = request.server_port

    # Convert ip address to network byte order
    ipv4 = struct.unpack('!i', socket.inet_aton(host))[0]
    port = int(port)
    # Zipkin passes unsigned values in signed types because Thrift has no
    # unsigned types, so we have to convert the value.
    port = struct.unpack('h', struct.pack('H', port))[0]

    return zipkin_core.Endpoint(
        ipv4=ipv4, port=port, service_name=service_name)


def annotation_list_builder(annotations, host):
    """
    Reformat annotations dict to return list of corresponding zipkin_core objects.

    :param annotations: dict containing key as annotation name,
                        value being timestamp in seconds(float).
    :type host: :class:`zipkin_core.Endpoint`
    :returns: a list of annotation zipkin_core objects
    :rtype: list
    """
    return [create_annotation(
        int(timestamp * 1000000), key, host)
        for key, timestamp in annotations.items()]


def binary_annotation_list_builder(binary_annotations, host):
    """
    Reformat binary annotations dict to return list of zipkin_core objects. The
    value of the binary annotations MUST be in string format.

    :param binary_annotations: dict with key, value being the name and value
                               of the binary annotation being logged.
    :type host: :class:`zipkin_core.Endpoint`
    :returns: a list of binary annotation zipkin_core objects
    :rtype: list
    """
    # TODO: Remove the type hard-coding of STRING to take it as a param option.
    ann_type = zipkin_core.AnnotationType.STRING
    return [create_binary_annotation(
        key, value, ann_type, host)
        for key, value in binary_annotations.items()]


def create_span(zipkin_attrs, span_name, annotations, binary_annotations,
                is_client=False):
    """
    Creates a zipkin span object based on a request

    :param request: pyramid request object
    :param annotations: list of zipkin annotation objects
    :param binary_annotations: list of zipkin binary annotation objects
    :returns: zipkin span object
    """

    span_id = generate_span_id() if is_client else zipkin_attrs.span_id
    parent_span_id = (
        zipkin_attrs.span_id if is_client else zipkin_attrs.parent_span_id)

    return zipkin_core.Span(**{
        "trace_id": get_id(zipkin_attrs.trace_id),
        "name": span_name,
        "id": get_id(span_id),
        "parent_id": get_id(parent_span_id),
        "annotations": annotations,
        "binary_annotations": binary_annotations,
    })


def thrift_obj_in_bytes(thrift_obj):  # pragma: no cover
    """
    Returns TBinaryProtocol encoded Thrift object.

    :param thrift_obj: thrift object to encode
    :returns: thrift object in TBinaryProtocol format bytes.
    """
    trans = TMemoryBuffer()
    thrift_obj.write(TBinaryProtocol(trans))

    return bytes(trans.getvalue())


def get_id(id_value):
    """
    Get the integer version of an encoded zipkin ID

    :param id_value: the string representation of a zipkin ID
    :returns: int representation of zipkin ID
    """
    if len(id_value) == 0:
        return 0

    return int(id_value, 16)



# -*- coding: utf-8 -*-

import codecs
import os
import re
import struct
from collections import namedtuple

from pyramid.interfaces import IRoutesMapper


DEFAULT_REQUEST_TRACING_PERCENT = 0.5


class ZipkinAttrs(namedtuple(
        'ZipkinAttrs', 'trace_id span_id parent_span_id flags is_sampled')):
    """
    Holds the basic attributes needed to log a zipkin trace

    :param trace_id: Unique trace id
    :param span_id: Span Id of the current request span
    :param parent_span_id: Parent span Id of the current request span
    :param flags: stores flags header. Currently unused
    :param is_sampled: pre-computed boolean whether the trace should be logged
    """


def generate_random_64bit_string():
    """Returns a 64 bit ascii encoded string"""
    return codecs.encode(os.urandom(8), 'hex_codec')


def thrift_compatble_string(token_id):
    """Converts token to a thrift compatible 64bit string"""
    # Zipkin passes unsigned values in signed types because Thrift has no
    # unsigned types, so we have to convert the value.
    return str(hex(struct.unpack('q', struct.pack('Q', int(token_id, 16)))[0]))


def generate_span_id():
    """
    Generates a new random span id

    :returns: string representation of zipkin span id
    """
    return thrift_compatble_string(generate_random_64bit_string())


def get_trace_id(request):
    """Gets the trace id based on a request. If not present with the request,
    create a custom (depending on config: `zipkin.trace_id_generator`) or a
    completely random trace id.

    :param: current active pyramid request
    :returns: string representaiton of zipkin trace id
    """
    if 'X-B3-TraceId' in request.headers:
        return request.headers['X-B3-TraceId']
    elif 'zipkin.trace_id_generator' in request.registry.settings:
        return request.registry.settings['zipkin.trace_id_generator'](request)
    else:
        return thrift_compatble_string(generate_random_64bit_string())


def should_not_sample_path(request):
    """Decided whether current request path should be sampled or not. This is
    checked previous to `should_not_sample_route` and takes precedence.

    :param: current active pyramid request
    :returns: boolean whether current request path is blacklisted.
    """
    blacklisted_paths = request.registry.settings.get(
        'zipkin.blacklisted_paths', [])
    regexes = [re.compile(r) for r in blacklisted_paths]
    return any(r.match(request.path) for r in regexes)


def should_not_sample_route(request):
    """Decided whether current request route should be sampled or not.

    :param: current active pyramid request
    :returns: boolean whether current request route is blacklisted.
    """
    blacklisted_routes = request.registry.settings.get(
        'zipkin.blacklisted_routes', [])

    if not blacklisted_routes:
        return False
    route_mapper = request.registry.queryUtility(IRoutesMapper)
    route_info = route_mapper(request).get('route')
    return (route_info and route_info.name in blacklisted_routes)


def should_sample_as_per_zipkin_tracing_percent(tracing_percent, req_id):
    """Calculate whether the request should be traced as per tracing percent.

    :param tracing_percent: value between 0.0 to 100.0
    :type tracing_percent: float
    :param req_id: unique request id of the request
    :returns: boolean whether current request should be sampled.
    """
    if tracing_percent == 0.0:  # Prevent the ZeroDivision
        return False
    inverse_frequency = int((1.0 / tracing_percent) * 100)
    return int(req_id, 16) % inverse_frequency == 0


def is_tracing(request):
    """Determine if zipkin should be tracing
    1) Check whether the current request path is blacklisted.
    2) If not, check whether the current request route is blacklisted.
    3) If not, check if specific sampled header is present in the request.
    4) If not, Use a tracing percent (default: 0.5%) to decide.

    :param request: pyramid request object

    :returns: boolean True if zipkin should be tracing
    """
    if should_not_sample_path(request):
        return False
    elif should_not_sample_route(request):
        return False
    elif 'X-B3-Sampled' in request.headers:
        return request.headers.get('X-B3-Sampled') == '1'
    else:
        zipkin_tracing_percent = request.registry.settings.get(
            'zipkin.tracing_percent', DEFAULT_REQUEST_TRACING_PERCENT)
        return should_sample_as_per_zipkin_tracing_percent(
            zipkin_tracing_percent, request.zipkin_trace_id)


def create_zipkin_attr(request):
    """Create ZipkinAttrs object from a request with sampled flag as True.
    Attaches lazy attribute `zipkin_trace_id` with request which is then used
    throughout the tween.

    :param request: pyramid request object
    :rtype: :class:`pyramid_zipkin.request_helper.ZipkinAttrs`
    """
    request.set_property(get_trace_id, 'zipkin_trace_id', reify=True)

    trace_id = request.zipkin_trace_id
    is_sampled = is_tracing(request)
    span_id = request.headers.get('X-B3-SpanId', '1')
    parent_span_id = request.headers.get('X-B3-ParentSpanId', '0')
    flags = request.headers.get('X-B3-Flags', '0')
    return ZipkinAttrs(trace_id=trace_id, span_id=span_id,
                       parent_span_id=parent_span_id,
                       flags=flags, is_sampled=is_sampled)



# -*- coding: utf-8 -*-

import logging
import time

from pyramid_zipkin.exception import ZipkinError
from pyramid_zipkin.thread_local import pop_zipkin_attrs
from pyramid_zipkin.thread_local import push_zipkin_attrs
from pyramid_zipkin.thrift_helper import annotation_list_builder
from pyramid_zipkin.thrift_helper import binary_annotation_list_builder
from pyramid_zipkin.thrift_helper import create_span
from pyramid_zipkin.thrift_helper import thrift_obj_in_bytes


zipkin_logger = logging.getLogger('pyramid_zipkin.logger')
zipkin_logger.setLevel(logging.DEBUG)


class ZipkinLoggingContext(object):
    """The main logging context manager which controls logging handler and
    stores the zipkin attributes on its creation.

    :type zipkin_attrs: :class:`pyramid_zipkin.request_helper.ZipkinAttrs`
    :type endpoint_attrs: :class:`pyramid_zipkin.zipkinCore.ttypes.Endpoint`
    :param log_handler: log handler to be attached to the module logger.
    :type log_handler: :class:`pyramid_zipkin.logging_helper.ZipkinLoggerHandler`
    :param request: active pyramid request object
    """
    def __init__(self, zipkin_attrs, endpoint_attrs, log_handler, request):
        self.zipkin_attrs = zipkin_attrs
        self.endpoint_attrs = endpoint_attrs
        self.handler = log_handler
        self.request_path = request.path
        self.request_path_qs = request.path_qs
        self.request_method = request.method
        self.registry_settings = request.registry.settings
        self.response_status_code = 0

    def __enter__(self):
        """Actions to be taken before request is handled.
        1) Attach `zipkin_logger` to :class:`ZipkingLoggerHandler` object.
        2) Push zipkin attributes to thread_local stack.
        3) Record the start timestamp.
        """
        zipkin_logger.addHandler(self.handler)
        push_zipkin_attrs(self.zipkin_attrs)
        self.start_timestamp = time.time()
        return self

    def __exit__(self, _type, _value, _traceback):
        """Actions to be taken post request handling.
        1) Record the end timestamp.
        2) Pop zipkin attributes from thread_local stack
        3) Detach `zipkin_logger` handler.
        4) And finally, if sampled, log the service annotations to scribe
        """
        self.log_spans()
        pop_zipkin_attrs()
        zipkin_logger.removeHandler(self.handler)

    def is_response_success(self):
        """Returns a boolean whether the response was a success
        """
        return 200 <= self.response_status_code <= 299

    def log_spans(self):
        """Main function to log all the annotations stored during the entire
        request. This is done if the request is sampled and the response was
        a success. It also logs the service `ss` and `sr` annotations.
        """
        if self.zipkin_attrs.is_sampled and self.is_response_success():
            for _, span in self.handler.spans.items():
                annotations = annotation_list_builder(
                    span['annotations'], self.endpoint_attrs)
                binary_annotations = binary_annotation_list_builder(
                    span['binary_annotations'], self.endpoint_attrs)
                log_span(self.zipkin_attrs, span['span_name'],
                         self.registry_settings, annotations,
                         binary_annotations, span['is_client'])

            end_timestamp = time.time()
            log_service_span(self.zipkin_attrs, self.start_timestamp,
                             end_timestamp, self.request_path,
                             self.request_path_qs, self.endpoint_attrs,
                             self.request_method, self.registry_settings)


class ZipkinLoggerHandler(logging.StreamHandler, object):
    """Logger Handler to log span annotations to scribe.
    To connect to the handler, logger name must be 'pyramid_zipkin.logger'

    :param zipkin_attrs: tuple containing trace_id, span_id & is_sampled
    """

    def __init__(self, zipkin_attrs):
        super(ZipkinLoggerHandler, self).__init__()
        self.zipkin_attrs = zipkin_attrs
        self.spans = {}

    def store_span(
            self, span_name, is_client, annotations, binary_annotations):
        """Store the annotations into a dict and send them later.

        :param span_name: string name of the span to be used.
        :param is_client: boolean to decide whether it is a client/server span
        :param annotations: dict of annotations logged.
        :param binary_annotations: dict of binary annotations logged.

        .. note::

            If duplicate annotations are logged for the same span name, only
            the last one will be logged.
        """
        key = (span_name, is_client)
        if key in self.spans:
            self.spans[key]['annotations'].update(annotations)
            self.spans[key]['binary_annotations'].update(binary_annotations)
        else:
            self.spans[key] = {'annotations': annotations,
                               'binary_annotations': binary_annotations,
                               'span_name': span_name,
                               'is_client': is_client
                               }

    def emit(self, record):
        """Handle each record message.

        :param record: object containing the `msg` object.
                       Structure of record.msg should be the following:
                       ::

                           {
                               "annotations": {
                                        "cs": ts1,
                                        "cr": ts2,
                                        },
                               "binary_annotations": {
                                        "http.uri": "/foo/bar",
                                        },
                                "name": "foo_span",
                                "type": "client/service",
                            }

                        "type" if not included is considered a 'service' span;
                        i.e. annotations are added in the current span instead
                        of creating a new one (which is the case for 'client').
        """
        if not self.zipkin_attrs.is_sampled:
            return
        span_name = record.msg.get('name', 'span')
        annotations = record.msg.get('annotations', {})
        binary_annotations = record.msg.get('binary_annotations', {})
        if not annotations and not binary_annotations:
            raise ZipkinError("Atleast one of annotation/binary annotation has"
                              " to be provided for {0} span".format(span_name))
        is_client = record.msg.get('type', 'service') == 'client'
        self.store_span(span_name, is_client, annotations, binary_annotations)


def log_span(zipkin_attrs, span_name, registry_settings, annotations,
             binary_annotations, is_client):
    """Creates a span and logs it.

    If `zipkin.transport_handler` config is set, it is used to act as a callback
    and log message is sent as a parameter.
    """
    span = create_span(
        zipkin_attrs, span_name, annotations, binary_annotations, is_client)
    message = thrift_obj_in_bytes(span)

    scribe_stream = registry_settings.get('zipkin.stream_name', 'zipkin')

    if 'zipkin.transport_handler' in registry_settings:
        return registry_settings['zipkin.transport_handler'](scribe_stream,
                                                             message)
    else:
        raise ZipkinError(
            "`zipkin.transport_handler` is a required config property, which"
            " is missing. It is a callback method which takes stream_name and"
            " a message as the params and logs message via scribe/kafka.")


def log_service_span(zipkin_attrs, start_timestamp, end_timestamp,
                     path, path_qs, endpoint, method, registry_settings):
    """Logs a span with `ss` and `sr` annotations.
    """
    annotations = annotation_list_builder(
        {'sr': start_timestamp, 'ss': end_timestamp}, endpoint)
    binary_annotations = binary_annotation_list_builder(
        {'http.uri': path, 'http.uri.qs': path_qs}, endpoint)
    log_span(zipkin_attrs, method, registry_settings,
             annotations, binary_annotations, False)



from pyramid.tweens import MAIN


def includeme(config):  # pragma: no cover
    """
    :type config: :class:`pyramid.config.Configurator`
    """
    config.add_tween('pyramid_zipkin.zipkin.zipkin_tween', over=MAIN)



#!/usr/bin/python
# -*- coding: utf-8 -*-

from setuptools import find_packages
from setuptools import setup

__version__ = "0.4.0"

setup(
    name='pyramid_zipkin',
    version=__version__,
    provides=["pyramid_zipkin"],
    author='Yelp, Inc.',
    author_email='opensource+pyramid-zipkin@yelp.com',
    license='Copyright Yelp 2015',
    url="https://github.com/Yelp/pyramid_zipkin",
    description='Zipkin distributed tracing system support library for pyramid.',
    packages=find_packages(exclude=('tests*', 'testing*', 'tools*')),
    package_data = { '': [ '*.thrift' ] },
    install_requires = [
        'pyramid',
        'thriftpy',
    ],
)
